function x = iterative_shrinkage_l2_batch(A, y, n, varargin)
% function x = iterative_shrinkage_l2_batch(A, y, n, varargin)
%
% Iterative shrinkage algorithm for sparse reconstruction
% Approach follows the methodology presented in the paper: Beck, Teboulle,
% A fast iterative shrinkage-thresholding algorithm for linear inverse
% problems, SIAM J. Imaging Sciences, 2009.
%
% Input: A - mxn matrix or function handle; in the case of function handle,
%        transpose operator should be provided
%
%        y - mxN matrix of right-side vectors
%
%        optional parameters: lambda - 1xN vector of regularization factors
%
%                             maxiter - max. nr of iterations
%
%                             tol - tolerance for stopping criterion
%
% Output: x - nxN matrix of solution vectors
%
% M. F., 09.03.2011.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Some parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
T = size(y, 2);

maxiter = 100;
maxinner = 50;
tol = 1e-6;
funcTol = 1e-6;
init = 'zeros';
verbose = false;

if isfloat(A)
    D = @(x) A*x;
    Dtrans = @(x) A'*x;
elseif isa(A, 'function_handle') 
    D = A;    
else
    error('A is of wrong type!');
end

% Initial value of L
L = ones(1, T);

% Check for optional parameters
if(mod(length(varargin),2) == 1)
    error('Optional arguments come in pairs.');
end
for i = 1:2:length(varargin)
    switch lower(varargin{i})
        case 'lambda'
            % lambda is a vector
            lambda = varargin{i+1};
        case 'maxiter'
            maxiter = varargin{i+1};
        case 'maxinner'
            maxinner = varargin{i+1};
        case 'atrans'
            Dtrans = varargin{i+1};
        case 'l'
            L = varargin{i+1};
        case 'tol'
            tol = varargin{i+1};   
        case 'init'
            init = varargin{i+1};
        case 'x0'
            x = varargin{i+1};
        case 'verbose'
            verbose = varargin{i+1};
    end
end

% Check if Atrans is provided
if ~exist('Dtrans', 'var')
    error('Transpose operator not provided!');
end

% Regularization constant
if ~exist('lambda', 'var')    
    lambda = 2e-2 * max(abs(Dtrans(y)));
end

% Check size of lambda
if (size(lambda, 2) ~= T)
    error('lambda should be a vector.');
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% L update factor (we increase L in every iteration if neccessary)
eta = 1.5;

% Initialization
if ~exist('x', 'var')    
    switch lower(init)
        case 'zeros'            
            x = zeros(n, T);
        case 'pinv'
            % If D is a tight frame, pseudoinverse corresponds to transpose
            x = Dtrans(y);
    end
end

xOld = x;
znj = x;
func_value_old = Inf(1, T);
tOld = 1;

diff = Inf(1, T);
funcDiff = Inf(1, T);
iter = 0;

if verbose
    fprintf('iter_nr    residual-norm    l1-norm    shrink-factor    inner-iter\n');
end

while any(diff > tol) && any(funcDiff > funcTol) && iter < maxiter
        
    residual = D(znj) - y;    
        
    l2_Old = sum(residual.^2) / 2;    
    grad = Dtrans(residual);         
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Backtracking    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
    step = L;
    shrink_factor = lambda ./ step;    
    znjUpdate = znj - grad ./ repmat(step, n, 1);
    x = sign(znjUpdate) .* ...
        max(abs(znjUpdate) - repmat(shrink_factor, n, 1), 0);            
    
    % Calculate value of quadratic approximation
    % Precompute some variables
    residual = D(x) - y;
        
    xnorm = sum(abs(x));    
    aux = l2_Old + sum((x-znj) .* grad) + lambda .* xnorm;
        
    func_approx = aux + step .* sum((x-znj).^2) / 2;
    
    % Calculate function value    
    l2 = sum(residual.^2) / 2;
    func_value = l2 + lambda .* xnorm;
    
    inner_iter = 1;
    Indices = (func_value > func_approx);    
    while any(Indices) && ...
            inner_iter <= maxinner           
        
        inner_iter = inner_iter+1;        
        step(Indices) = (eta ^ inner_iter) * L(Indices);
        shrink_factor = lambda ./ step;
                
        znjUpdate(:, Indices) = znj(:, Indices) - ...
            grad(:, Indices) ./ repmat(step(Indices), n, 1);
        x(:, Indices) = sign(znjUpdate(:, Indices)) .* ...
            max(abs(znjUpdate(:, Indices)) - ...
            repmat(shrink_factor(Indices), n, 1), 0);
         
        % Calculate value of quadratic approximation
        % Precompute some variables
        residual = D(x) - y;    
        
        xnorm = sum(abs(x));
        
        aux = l2_Old + sum((x-znj) .* grad) + lambda .* xnorm;
        
        func_approx = aux + step .* sum((x-znj).^2) / 2;
        
        % Select Indices where backtracking should be done
        Indices = (func_value > func_approx);

        % Calculate function value        
        l2 = sum(residual.^2) / 2;
        func_value = l2 + lambda .* xnorm;        
        
    end
    L = step;
    
    t = (1 + sqrt(1 + 4*tOld^2)) / 2;
    znj = x + (tOld-1)/t * (x - xOld);
    tOld = t;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
    diff = max(abs(x-xOld));
    funcDiff = abs(func_value_old - func_value);
    func_value_old = func_value;
    xOld = x;
    
    iter = iter+1;
    
    if verbose
        fprintf('   %d    %2.10e     %2.10e    %2.1e    %d\n', iter, ...
            max(sum(residual.^2)), ...
            max(sum(abs(x))), min(shrink_factor), inner_iter-1);
    end
end
